package idtools

const (
	SeTakeOwnershipPrivilege = "SeTakeOwnershipPrivilege"
)

// TODO(thaJeztah): these magic consts need a source of reference, and should be defined in a canonical location
const (
	ContainerAdministratorSidString = "S-1-5-93-2-1"

	ContainerUserSidString = "S-1-5-93-2-2"
)
